import numpy as np
import functools
from fastapi import Request, BackgroundTasks
from typing import Any, Dict, Optional, Union

import supervisely as sly
import supervisely.nn.inference.tracking.functional as F
from supervisely.annotation.label import Label
from supervisely.nn.prediction_dto import PredictionSegmentation
from supervisely.nn.inference.tracking.tracker_interface import TrackerInterface
from supervisely.nn.inference.cache import InferenceImageCache
from supervisely.nn.inference import Inference


class MaskTracking(Inference, InferenceImageCache):
    def __init__(
        self,
        model_dir: Optional[str] = None,
        custom_inference_settings: Optional[Union[Dict[str, Any], str]] = None,
    ):
        Inference.__init__(
            self,
            model_dir,
            custom_inference_settings,
            sliding_window_mode=None,
            use_gui=False,
        )
        InferenceImageCache.__init__(
            self,
            maxsize=sly.env.smart_cache_size(),
            ttl=sly.env.smart_cache_ttl(),
            is_persistent=True,
        )

        try:
            self.load_on_device(model_dir, "cuda")
        except RuntimeError:
            self.load_on_device(model_dir, "cpu")
            sly.logger.warn("Failed to load model on CUDA device.")

        sly.logger.debug(
            "Smart cache params",
            extra={
                "ttl": sly.env.smart_cache_ttl(),
                "maxsize": sly.env.smart_cache_size(),
            },
        )

    def get_info(self):
        info = super().get_info()
        info["task type"] = "tracking"
        return info

    def serve(self):
        super().serve()
        server = self._app.get_server()
        self.add_cache_endpoint(server)

        @server.post("/track")
        def start_track(request: Request, task: BackgroundTasks):
            task.add_task(track, request)
            return {"message": "Track task started."}

        def send_error_data(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                value = None
                try:
                    value = func(*args, **kwargs)
                except Exception as exc:
                    # print("An error occured:")
                    # print(traceback.format_exc())
                    request: Request = args[0]
                    context = request.state.context
                    api: sly.Api = request.state.api
                    track_id = context["trackId"]
                    api.logger.error("An error occured:")
                    api.logger.exception(exc)

                    api.post(
                        "videos.notify-annotation-tool",
                        data={
                            "type": "videos:tracking-error",
                            "data": {
                                "trackId": track_id,
                                "error": {"message": repr(exc)},
                            },
                        },
                    )
                return value

            return wrapper

        @send_error_data
        def track(request: Request):
            context = request.state.context
            api: sly.Api = request.state.api

            self.video_interface = TrackerInterface(
                context=context,
                api=api,
                load_all_frames=True,
                notify_in_predict=True,
                per_point_polygon_tracking=False,
                frame_loader=self.download_frame,
            )
            api.logger.info("Starting tracking process")
            # load frames
            frames = self.video_interface.frames
            # combine several binary masks into one multilabel mask
            i = 0
            label2id = {}

            for (fig_id, geometry), obj_id in zip(
                self.video_interface.geometries.items(),
                self.video_interface.object_ids,
            ):
                original_geometry = geometry.clone()
                if not isinstance(geometry, sly.Bitmap) and not isinstance(
                    geometry, sly.Polygon
                ):
                    raise TypeError(
                        f"This app does not support {geometry.geometry_name()} tracking"
                    )
                # convert polygon to bitmap
                if isinstance(geometry, sly.Polygon):
                    polygon_obj_class = sly.ObjClass("polygon", sly.Polygon)
                    polygon_label = sly.Label(geometry, polygon_obj_class)
                    bitmap_obj_class = sly.ObjClass("bitmap", sly.Bitmap)
                    bitmap_label = polygon_label.convert(bitmap_obj_class)[0]
                    geometry = bitmap_label.geometry
                if i == 0:
                    multilabel_mask = geometry.data.astype(int)
                    multilabel_mask = np.zeros(frames[0].shape, dtype=np.uint8)
                    geometry.draw(bitmap=multilabel_mask, color=[1, 1, 1])
                    i += 1
                else:
                    i += 1
                    geometry.draw(bitmap=multilabel_mask, color=[i, i, i])
                label2id[i] = {
                    "fig_id": fig_id,
                    "obj_id": obj_id,
                    "original_geometry": original_geometry.geometry_name(),
                }
            # run tracker
            tracked_multilabel_masks = self.predict(
                frames=frames, input_mask=multilabel_mask[:, :, 0]
            )
            tracked_multilabel_masks = np.array(tracked_multilabel_masks)
            # decompose multilabel masks into binary masks
            for i in np.unique(tracked_multilabel_masks):
                if i != 0:
                    binary_masks = tracked_multilabel_masks == i
                    fig_id = label2id[i]["fig_id"]
                    obj_id = label2id[i]["obj_id"]
                    geometry_type = label2id[i]["original_geometry"]
                    for j, mask in enumerate(binary_masks[1:]):
                        # check if mask is not empty
                        if not np.any(mask):
                            api.logger.info(
                                f"Skipping empty mask on frame {self.video_interface.frame_index + j + 1}"
                            )
                            # update progress bar anyway (otherwise it will not be finished)
                            self.video_interface._notify(task="add geometry on frame")
                        else:
                            if geometry_type == "polygon":
                                bitmap_geometry = sly.Bitmap(mask)
                                bitmap_obj_class = sly.ObjClass("bitmap", sly.Bitmap)
                                bitmap_label = sly.Label(
                                    bitmap_geometry, bitmap_obj_class
                                )
                                polygon_obj_class = sly.ObjClass("polygon", sly.Polygon)
                                polygon_labels = bitmap_label.convert(polygon_obj_class)
                                geometries = [
                                    label.geometry for label in polygon_labels
                                ]
                            else:
                                geometries = [sly.Bitmap(mask)]
                            for l, geometry in enumerate(geometries):
                                if l == len(geometries) - 1:
                                    notify = True
                                else:
                                    notify = False
                                self.video_interface.add_object_geometry_on_frame(
                                    geometry,
                                    obj_id,
                                    self.video_interface.frames_indexes[j + 1],
                                    notify=notify,
                                )
                    if self.video_interface.global_stop_indicatior:
                        return
                    api.logger.info(f"Figure with id {fig_id} was successfully tracked")



from functools import wraps
from json import JSONDecodeError
from requests.exceptions import HTTPError, RetryError
from shutil import ReadError
from typing import List, Union, Callable, Dict
from rich.console import Console

from supervisely.sly_logger import logger, EventType
from supervisely.app import DialogWindowError

import traceback
import re

# TODO: Add correct doc link.
# DOC_URL = "https://docs.supervisely.com/errors/"


class HandleException:
    def __init__(
        self,
        exception: Exception,
        stack: List[traceback.FrameSummary] = None,
        **kwargs,
    ):
        self.exception = exception
        self.stack = stack or read_stack_from_exception(self.exception)
        self.code = kwargs.get("code")
        self.title = kwargs.get("title")

        # TODO: Add error code to the title.
        # self.title += f" (SLYE{self.code})"

        self.message = kwargs.get("message")

        # TODO: Add doc link to the message.
        # error_url = DOC_URL + str(self.code)
        # self.message += (
        #     f"<br>Check out the <a href='{error_url}'>documentation</a> "
        #     "for solutions and troubleshooting."
        # )

        self.dev_logging()

    def dev_logging(self):
        # * Option for logging the exception with sly logger.
        # * Due to JSON formatting it's not convinient to read the log.
        # tracelog = " | ".join(traceback.format_list(self.stack))
        # logger.error(
        #     "Detailed info about exception."
        #     f"Exception type: {type(self.exception)}. "
        #     f"Exception value: {self.exception}. "
        #     f"Error code: {self.code}. "
        #     f"Error title: {self.title}. "
        #     f"Error message: {self.message} "
        #     f"Traceback: {tracelog}. "
        # )

        # * Printing the exception to the console line by line.
        console = Console()

        console.print("🛑 Beginning of the error report.", style="bold red")
        console.print(f"🔴 {self.exception.__class__.__name__}: {self.exception}")
        # TODO: Uncomment code line when error codes will be added.
        # console.print(f"Error code: {self.code}.", style="bold orange")
        console.print(f"🔴 Error title: {self.title}")
        console.print(f"🔴 Error message: {self.message}")

        console.print("🔴 Traceback (most recent call last):", style="bold red")

        for i, trace in enumerate(traceback.format_list(self.stack)):
            console.print(f"{i + 1}. {trace}")
        console.print("🛑 End of the error report.", style="bold red")

    def raise_error(self):
        raise DialogWindowError(self.title, self.message)

    def log_error_for_agent(self, main_name: str):
        logger.critical(
            self.title,
            exc_info=True,
            extra={
                "main_name": main_name,
                "event_type": EventType.TASK_CRASHED,
                "exc_str": self.message,
            },
        )

    def get_message_for_modal_window(self):
        return f"{self.title}. \n{self.message}"


class ErrorHandler:
    class APP:
        class UnsupportedShapes(HandleException):
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary]):
                self.code = 1001
                self.title = "Unsupported class shapes"
                self.message = exception.args[0]

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class UnsupportedArchiveFormat(HandleException):
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary]):
                self.code = 1002
                self.title = "Unsupported archive format"
                self.message = "The given archive format is not supported."

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class UnicodeDecodeError(HandleException):
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary]):
                self.code = 1003
                self.title = "Unicode decode error"
                self.message = (
                    "The given file contains non-unicode characters. "
                    "Please, check the file and try again."
                )

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

    class API:
        class TeamFilesFileNotFound(HandleException):
            def __init__(
                self,
                exception: Exception,
                stack: List[traceback.FrameSummary] = None,
            ):
                self.code = 2001
                self.title = "Requested file was not found in Team Files"
                self.message = (
                    "The requested file doesn't exist in Team Files in current Team. "
                    "Please, ensure that the file exists and you're working in the correct Team."
                )

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class TaskSendRequestError(HandleException):
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary]):
                self.code = 2002
                self.title = "Task send request error"
                self.message = (
                    "The application has encountered an error while sending a request to the task. "
                    "Please, check that the task is running."
                )

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class FileSizeTooLarge(HandleException):
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary]):
                self.code = 2003
                self.title = "File size limit exceeded"
                self.message = (
                    "The given file size is too large (more than 10 GB) for free community edition."
                )

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class ImageFilesSizeTooLarge(HandleException):
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary]):
                self.code = 2004
                self.title = "Image files size limit exceeded"
                self.message = "The given image file size is too large (more than 1 GB) for free community edition."

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class VideoFilesSizeTooLarge(HandleException):
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary]):
                self.code = 2005
                self.title = "Video files size limit exceeded"
                self.message = "The given video file size is too large (more than 300 MB) for free community edition."

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class VolumeFilesSizeTooLarge(HandleException):
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary]):
                self.code = 2006
                self.title = "Volume files size limit exceeded"
                self.message = "The given volume file size is too large (more than 150 MB) for free community edition."

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class OutOfMemory(HandleException):
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary]):
                self.code = 2007
                self.title = "Out of memory"
                self.message = (
                    "The agent ran out of memory. "
                    "Please, check your agent's memory usage, reduce batch size or use a device with more memory capacity."
                )

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class DockerRuntimeError(HandleException):
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary]):
                self.code = 2008
                self.title = "Docker runtime error"
                self.message = (
                    "The agent has encountered a Docker runtime error. "
                    "Please, check that docker is installed and running."
                )

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class AppSetFieldError(HandleException):
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary]):
                self.code = 2009
                self.title = "App set field error"
                self.message = (
                    "The application has encountered an error while sending a request. "
                    "It may be caused by connection issues. "
                    "Please, check your internet connection, the app is running and input parameters. "
                    "If you are using a model in other session, please, check that the model is serving and it's logs."
                )

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class TeamFilesDirectoryDownloadError(HandleException):
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary]):
                self.code = 2010
                self.title = "Team files directory download error"
                self.message = "Make sure that the directory exists in the team files, the files are not corrupted, and try again."

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class PointcloudsUploadError(HandleException):
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary]):
                self.code = 2011
                self.title = "Pointclouds uploading error"
                self.message = exception.args[0]

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class AnnotationUploadError(HandleException):
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary]):
                self.code = 2012
                self.title = "Annotation uploading error"
                self.message = exception.args[0]

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class AnnotationNotFound(HandleException):
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary]):
                self.code = 2013
                self.title = "Annotation not found"
                self.message = "Please, check that the annotation(s) exists by the given path."

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class ServerOverload(HandleException):
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary]):
                self.code = 2014
                self.title = "High load on the server"
                self.message = "Sorry, the server is overloaded. Please, try again later."

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message
                )

        class ProjectNotFound(HandleException):
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary]):
                self.code = 2015
                self.title = "Project not found"
                self.message = "Please, check that the project exists, not archived and you have enough permissions to access it"

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class DatasetNotFound(HandleException):
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary]):
                self.code = 2016
                self.title = "Dataset not found"
                self.message = "Please, check that the dataset exists, not archived and you have enough permissions to access it."

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

    class SDK:
        class ProjectStructureError(HandleException):
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary]):
                self.code = 3001
                self.title = "Project structure error"
                self.message = "Please, check the project structure and try again."

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class ConversionNotImplemented(HandleException):
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary]):
                self.code = 3002
                self.title = "Not implemented error"
                self.message = (
                    "Conversion is not implemented between the given object types. "
                    "Please, check the geometry type of the objects."
                )

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class JsonAnnotationReadError(HandleException):
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary]):
                self.code = 3003
                self.title = "JSON annotation read error"
                self.message = (
                    "Please, check that the file has the correct Supervisely JSON format."
                )

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )

        class LabelFromJsonFailed(HandleException):
            def __init__(self, exception: Exception, stack: List[traceback.FrameSummary]):
                self.code = 3004
                self.title = "Label deserialize error"
                self.message = exception.args[0]

                super().__init__(
                    exception,
                    stack,
                    code=self.code,
                    title=self.title,
                    message=self.message,
                )


ERROR_PATTERNS = {
    AttributeError: {r".*api\.file\.get_info_by_path.*": ErrorHandler.API.TeamFilesFileNotFound},
    HTTPError: {
        r".*api\.task\.send_request.*": ErrorHandler.API.TaskSendRequestError,
        r".*api\.app\.set_field.*": ErrorHandler.API.AppSetFieldError,
        r".*file-storage\.bulk\.upload.*File too large.*": ErrorHandler.API.FileSizeTooLarge,
        r".*images\.bulk\.upload.*FileSize.*\"sizeLimit\":1073741824.*": ErrorHandler.API.ImageFilesSizeTooLarge,
        r".*videos\.bulk\.upload.*FileSize.*sizeLimit\":314572800.*": ErrorHandler.API.VideoFilesSizeTooLarge,
        r".*images\.bulk\.upload.*FileSize.*\"sizeLimit\":157286400.*": ErrorHandler.API.VolumeFilesSizeTooLarge,
        r".*Dataset with datasetId.*is either archived, doesn't exist or you don't have enough permissions to access.*": ErrorHandler.API.DatasetNotFound,
        r".*Project with projectId.*is either archived, doesn't exist or you don't have enough permissions to access.*": ErrorHandler.API.ProjectNotFound,
    },
    RuntimeError: {
        r".*Label\.from_json.*": ErrorHandler.SDK.LabelFromJsonFailed,
    },
    FileNotFoundError: {
        r".*api\.annotation\.upload_path.*": ErrorHandler.API.AnnotationNotFound,
        r".*api\.annotation\.upload_paths.*": ErrorHandler.API.AnnotationNotFound,
        r".*read_single_project.*": ErrorHandler.SDK.ProjectStructureError,
    },
    JSONDecodeError: {
        r".*sly\.json\.load_json_file.*": ErrorHandler.SDK.JsonAnnotationReadError,
        r".*api\.annotation\.upload_path.*": ErrorHandler.SDK.JsonAnnotationReadError,
        r".*api\.annotation\.upload_paths.*": ErrorHandler.SDK.JsonAnnotationReadError,
    },
    UnicodeDecodeError: {
        r".*codec can't decode byte.*": ErrorHandler.APP.UnicodeDecodeError,
    },
    NotImplementedError: {
        r".*geometry\.convert.*": ErrorHandler.SDK.ConversionNotImplemented,
    },
    ValueError: {
        r".*obj_class\.geometry_type\.from_json.*": ErrorHandler.SDK.LabelFromJsonFailed,
    },
    IndexError: {
        r".*obj_class\.geometry_type\.from_json.*": ErrorHandler.SDK.LabelFromJsonFailed,
    },
    ReadError: {
        r".*shutil\.unpack_archive.*": ErrorHandler.APP.UnsupportedArchiveFormat,
        r".*api\.file\.download_directory.*": ErrorHandler.API.TeamFilesDirectoryDownloadError,
    },
    KeyError: {
        r".*api\.pointcloud\.upload_paths.*": ErrorHandler.API.PointcloudsUploadError,
        r".*api\.pointcloud\.upload_project.*": ErrorHandler.SDK.ProjectStructureError,
        r".*api\.image\.download_bytes.*": ErrorHandler.API.ServerOverload,
        r".*api\.video\.frame\.download_np.*": ErrorHandler.API.ServerOverload,
        r".*api\.image\.download_bytes.*": ErrorHandler.API.ServerOverload,
    },
    TypeError: {
        r".*obj_class\.geometry_type\.from_json.*": ErrorHandler.SDK.LabelFromJsonFailed,
    },
    RetryError: {
        r".*api\.annotation\.upload_paths.*": ErrorHandler.API.AnnotationUploadError,
        r".*api\.task\.set_field.*": ErrorHandler.API.AppSetFieldError,
    },
    RuntimeError: {r".*CUDA out of memory.*Tried to allocate.*": ErrorHandler.API.OutOfMemory},
    # Exception: {r".*unable to start container process.*": ErrorHandler.API.DockerRuntimeError},
}


def handle_exception(exception: Exception) -> Union[ErrorHandler, None]:
    """Function for handling exceptions, using the stack trace and patterns for known errors.
    Returns an instance of the ErrorHandler class if the pattern is found, otherwise returns None.

    :param exception: Exception to be handled.
    :type exception: Exception
    :return: Instance of the ErrorHandler class or None.
    :rtype: Union[ErrorHandler, None]
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        try:
            # Some code that may raise an exception.
        except Exception as e:
            exception_handler = sly.handle_exception(e)
            if exception_handler:
                # You may raise the exception using the raise_error() method.
                # Or use other approaches for handling the exception (e.g. logging, printing to console, etc.)
                exception_handler.raise_error()
            else:
                # If the pattern is not found, the exception is raised as usual.
                raise
    """
    # Extracting the stack trace.
    stack = read_stack_from_exception(exception)

    # Retrieving the patterns for the given exception type.
    patterns = ERROR_PATTERNS.get(type(exception))
    if not patterns:
        return

    # Looping through the stack trace from the bottom up to find matching pattern with specified Exception type.
    for pattern, handler in patterns.items():
        for frame in stack[::-1]:
            if re.match(pattern, frame.line):
                return handler(exception, stack)
        if re.match(pattern, exception.args[0]):
            return handler(exception, stack)


def handle_exceptions(func: Callable) -> Callable:
    """Decorator for handling exceptions, which tries to find a matching pattern for known errors.
    If the pattern is found, the exception is handled according to the specified handler.
    Otherwise, the exception is raised as usual.

    :param func: Function to be decorated.
    :type func: Callable
    :return: Decorated function.
    :rtype: Callable
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        @sly.handle_exceptions
        def my_func():
            # Some code that may raise an exception.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            exception_handler = handle_exception(e)
            if exception_handler:
                exception_handler.raise_error()
            else:
                raise

    return wrapper


def handle_additional_exceptions(
    errors: Dict[BaseException, Dict[str, HandleException]] = {},
) -> Callable:
    ERROR_PATTERNS.update(errors)

    def inner(func: Callable) -> Callable:
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                exception_handler = handle_exception(e)
                if exception_handler:
                    exception_handler.raise_error()
                else:
                    raise

        return wrapper

    return inner


def read_stack_from_exception(exception):
    stack = traceback.extract_stack()
    stack.extend(traceback.extract_tb(exception.__traceback__))
    return stack



from typing import Dict, List, Any, Literal, Optional, Union
from typing_extensions import Literal

from supervisely import env as sly_env
from supervisely.geometry.bitmap import Bitmap
from supervisely.nn.prediction_dto import PredictionMask
from supervisely.annotation.label import Label
from supervisely.sly_logger import logger
from supervisely.nn.inference.inference import Inference
from supervisely.nn.inference.cache import InferenceImageCache
from supervisely.decorators.inference import process_image_sliding_window


class PromptableSegmentation(Inference, InferenceImageCache):
    def __init__(
        self,
        model_dir: Optional[str] = None,
        custom_inference_settings: Optional[
            Union[Dict[str, Any], str]
        ] = None,  # dict with settings or path to .yml file
        sliding_window_mode: Optional[Literal["basic", "advanced", "none"]] = "basic",
        use_gui: Optional[bool] = False,
    ):
        Inference.__init__(self, model_dir, custom_inference_settings, sliding_window_mode, use_gui)
        InferenceImageCache.__init__(
            self,
            maxsize=sly_env.smart_cache_size(),
            ttl=sly_env.smart_cache_ttl(),
            is_persistent=True,
        )
        logger.debug(
            "Smart cache params",
            extra={"ttl": sly_env.smart_cache_ttl(), "maxsize": sly_env.smart_cache_size()},
        )

    def get_info(self) -> dict:
        info = super().get_info()
        info["task type"] = "promptable segmentation"
        return info

    def _get_obj_class_shape(self):
        return Bitmap

    def _create_label(self, dto: PredictionMask):
        obj_class = self.model_meta.get_obj_class(dto.class_name)
        if obj_class is None:
            raise KeyError(
                f"Class {dto.class_name} not found in model classes {self.get_classes()}"
            )
        if not dto.mask.any():  # skip empty masks
            logger.debug(f"Mask of class {dto.class_name} is empty and will be skipped")
            return None
        geometry = Bitmap(dto.mask)
        label = Label(geometry, obj_class)
        return label

    def predict(self, image_path: str, settings: Dict[str, Any]) -> List[PredictionMask]:
        raise NotImplementedError("Have to be implemented in child class")

    def predict_raw(self, image_path: str, settings: Dict[str, Any]) -> List[PredictionMask]:
        raise NotImplementedError(
            "Have to be implemented in child class If sliding_window_mode is 'advanced'."
        )

    @process_image_sliding_window
    def _inference_image_path(
        self,
        image_path: str,
        settings: Dict,
        data_to_return: Dict,  # for decorators
    ):
        inference_mode = settings.get("inference_mode", "full_image")
        logger.debug(
            "Inferring image_path:", extra={"inference_mode": inference_mode, "path": image_path}
        )

        if inference_mode == "sliding_window" and settings["sliding_window_mode"] == "advanced":
            predictions = self.predict_raw(image_path=image_path, settings=settings)
        else:
            predictions = self.predict(image_path=image_path, settings=settings)
        ann = self._predictions_to_annotation(image_path, predictions)

        logger.debug(
            f"Inferring image_path done. pred_annotation:",
            extra=dict(w=ann.img_size[1], h=ann.img_size[0], n_labels=len(ann.labels)),
        )
        return ann



# coding: utf-8

from supervisely.video_annotation.video_object import VideoObject


class VolumeObject(VideoObject):
    """
    VolumeObject object for :class:`VolumeAnnotation<supervisely.volume_annotation.volume_annotation.VolumeAnnotation>`. :class:`VolumeObject<VolumeObject>` object is immutable.

    :param obj_class: VolumeObject :class:`class<supervisely.annotation.obj_class.ObjClass>`.
    :type obj_class: ObjClass
    :param tags: VolumeObject :class:`tags<supervisely.volume_annotation.volume_tag_collection.VolumeTagCollection>`.
    :type tags: VolumeTagCollection, optional
    :param key: KeyIdMap object.
    :type key: KeyIdMap, optional
    :param class_id: ID of :class:`ObjClass<supervisely.annotation.obj_class.ObjClass>` to which VolumeObject belongs.
    :type class_id: int, optional
    :param labeler_login: Login of the user who created VolumeObject.
    :type labeler_login: str, optional
    :param updated_at: Date and Time when VolumeObject was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
    :type updated_at: str, optional
    :param created_at: Date and Time when VolumeObject was created. Date Format is the same as in "updated_at" parameter.
    :type created_at: str, optional
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        obj_class_heart = sly.ObjClass('heart', sly.Rectangle)
        volume_obj_heart = sly.VolumeObject(obj_class_heart)
        volume_obj_heart_json = volume_obj_heart.to_json()
        print(volume_obj_heart_json)
        # Output: {
        #     "key": "6b819f1840f84d669b32cdec225385f0",
        #     "classTitle": "heart",
        #     "tags": []
        # }
    """

    pass



import numpy as np
import functools
from fastapi import Request, BackgroundTasks
from typing import Any, Dict, List, Optional, Union
from pathlib import Path

import supervisely as sly
import supervisely.nn.inference.tracking.functional as F
from supervisely.annotation.label import Label
from supervisely.nn.prediction_dto import Prediction, PredictionBBox
from supervisely.nn.inference.tracking.tracker_interface import TrackerInterface
from supervisely.nn.inference import Inference
from supervisely.nn.inference.cache import InferenceImageCache


class BBoxTracking(Inference, InferenceImageCache):
    def __init__(
        self,
        model_dir: Optional[str] = None,
        custom_inference_settings: Optional[Union[Dict[str, Any], str]] = None,
    ):
        Inference.__init__(
            self,
            model_dir,
            custom_inference_settings,
            sliding_window_mode=None,
            use_gui=False,
        )
        InferenceImageCache.__init__(
            self,
            maxsize=sly.env.smart_cache_size(),
            ttl=sly.env.smart_cache_ttl(),
            is_persistent=True,
        )

        try:
            self.load_on_device(model_dir, "cuda")
        except RuntimeError:
            self.load_on_device(model_dir, "cpu")
            sly.logger.warn("Failed to load model on CUDA device.")

        sly.logger.debug(
            "Smart cache params",
            extra={"ttl": sly.env.smart_cache_ttl(), "maxsize": sly.env.smart_cache_size()},
        )

    def get_info(self):
        info = super().get_info()
        info["task type"] = "tracking"
        return info

    def serve(self):
        super().serve()
        server = self._app.get_server()
        self.add_cache_endpoint(server)

        @server.post("/track")
        def start_track(request: Request, task: BackgroundTasks):
            task.add_task(track, request)
            return {"message": "Track task started."}

        def send_error_data(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                value = None
                try:
                    value = func(*args, **kwargs)
                except Exception as exc:
                    request: Request = args[0]
                    context = request.state.context
                    api: sly.Api = request.state.api
                    track_id = context["trackId"]
                    api.logger.error("An error occured:")
                    api.logger.exception(exc)

                    api.post(
                        "videos.notify-annotation-tool",
                        data={
                            "type": "videos:tracking-error",
                            "data": {
                                "trackId": track_id,
                                "error": {"message": repr(exc)},
                            },
                        },
                    )
                return value

            return wrapper

        @send_error_data
        def track(request: Request):
            context = request.state.context
            api: sly.Api = request.state.api

            video_interface = TrackerInterface(
                context=context,
                api=api,
                load_all_frames=False,
                frame_loader=self.download_frame,
            )

            range_of_frames = [
                video_interface.frames_indexes[0],
                video_interface.frames_indexes[-1],
            ]

            self.run_cache_task_manually(
                api,
                [range_of_frames],
                video_id=video_interface.video_id,
            )

            api.logger.info("Start tracking.")

            for fig_id, obj_id in zip(
                video_interface.geometries.keys(),
                video_interface.object_ids,
            ):
                init = False
                for _ in video_interface.frames_loader_generator():
                    geom = video_interface.geometries[fig_id]
                    if not isinstance(geom, sly.Rectangle):
                        raise TypeError(f"Tracking does not work with {geom.geometry_name()}.")

                    imgs = video_interface.frames
                    target = PredictionBBox(
                        "",  # TODO: can this be useful?
                        [geom.top, geom.left, geom.bottom, geom.right],
                        None,
                    )

                    if not init:
                        self.initialize(imgs[0], target)
                        init = True

                    geometry = self.predict(
                        rgb_image=imgs[-1],
                        prev_rgb_image=imgs[0],
                        target_bbox=target,
                        settings=self.custom_inference_settings_dict,
                    )
                    sly_geometry = self._to_sly_geometry(geometry)
                    video_interface.add_object_geometries([sly_geometry], obj_id, fig_id)

                    if video_interface.global_stop_indicatior:
                        return

                api.logger.info(f"Figure #{fig_id} tracked.")

    def initialize(self, init_rgb_image: np.ndarray, target_bbox: PredictionBBox) -> None:
        """
        Initializing the tracker with a new object.

        :param init_rgb_image: frame with object
        :type init_rgb_image: np.ndarray
        :param target_bbox: initial bbox
        :type target_bbox: PredictionBBox
        """
        raise NotImplementedError

    def predict(
        self,
        rgb_image: np.ndarray,
        settings: Dict[str, Any],
        prev_rgb_image: np.ndarray,
        target_bbox: PredictionBBox,
    ) -> PredictionBBox:
        """
        SOT prediction

        :param rgb_image: search frame
        :type rgb_image: np.ndarray
        :param settings: model parameters
        :type settings: Dict[str, Any]
        :param init_rgb_image: previous frame with object
        :type init_rgb_image: np.ndarray
        :param target_bbox: bbox added on previous step
        :type target_bbox: PredictionBBox
        :return: predicted annotation
        :rtype: PredictionBBox
        """
        raise NotImplementedError

    def visualize(
        self,
        predictions: List[PredictionBBox],
        images: List[np.ndarray],
        vis_path: str,
        thickness: int = 2,
    ):
        vis_path = Path(vis_path)

        for i, (pred, image) in enumerate(zip(predictions, images)):
            out_path = vis_path / f"img_{i}.jpg"
            ann = self._predictions_to_annotation(image, [pred])
            ann.draw_pretty(
                bitmap=image,
                color=(255, 0, 0),
                thickness=thickness,
                output_path=str(out_path),
                fill_rectangles=False,
            )

    def _to_sly_geometry(self, dto: PredictionBBox) -> sly.Rectangle:
        top, left, bottom, right = dto.bbox_tlbr
        geometry = sly.Rectangle(top=top, left=left, bottom=bottom, right=right)
        return geometry

    def _create_label(self, dto: PredictionBBox) -> sly.Rectangle:
        geometry = self._to_sly_geometry(dto)
        return Label(geometry, sly.ObjClass("", sly.Rectangle))

    def _get_obj_class_shape(self):
        return sly.Rectangle

    def _predictions_to_annotation(
        self, image: np.ndarray, predictions: List[Prediction]
    ) -> sly.Annotation:
        labels = []
        for prediction in predictions:
            label = self._create_label(prediction)
            if label is None:
                # for example empty mask
                continue
            if isinstance(label, list):
                labels.extend(label)
                continue
            labels.append(label)

        # create annotation with correct image resolution
        ann = sly.Annotation(img_size=image.shape[:2])
        ann = ann.add_labels(labels)
        return ann



import numpy as np
import functools
from fastapi import Request, BackgroundTasks
from typing import Any, Dict, List, Optional, Union
from pathlib import Path

import supervisely as sly
import supervisely.nn.inference.tracking.functional as F
from supervisely.annotation.label import Label
from supervisely.nn.prediction_dto import Prediction, PredictionPoint
from supervisely.nn.inference.tracking.tracker_interface import TrackerInterface
from supervisely.nn.inference import Inference
from supervisely.nn.inference.cache import InferenceImageCache


class PointTracking(Inference, InferenceImageCache):
    def __init__(
        self,
        model_dir: Optional[str] = None,
        custom_inference_settings: Optional[Union[Dict[str, Any], str]] = None,
    ):
        Inference.__init__(
            self,
            model_dir,
            custom_inference_settings,
            sliding_window_mode=None,
            use_gui=False,
        )
        InferenceImageCache.__init__(
            self,
            maxsize=sly.env.smart_cache_size(),
            ttl=sly.env.smart_cache_ttl(),
            is_persistent=True,
        )

        try:
            self.load_on_device(model_dir, "cuda")
        except RuntimeError:
            self.load_on_device(model_dir, "cpu")
            sly.logger.warn("Failed to load model on CUDA device.")

        sly.logger.debug(
            "Smart cache params",
            extra={"ttl": sly.env.smart_cache_ttl(), "maxsize": sly.env.smart_cache_size()},
        )

    def get_info(self):
        info = super().get_info()
        info["task type"] = "tracking"
        # recommended parameters:
        # info["model_name"] = ""
        # info["checkpoint_name"] = ""
        # info["pretrained_on_dataset"] = ""
        # info["device"] = ""
        return info

    def serve(self):
        super().serve()
        server = self._app.get_server()
        self.add_cache_endpoint(server)

        @server.post("/track")
        def start_track(request: Request, task: BackgroundTasks):
            task.add_task(track, request)
            return {"message": "Track task started."}

        def send_error_data(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                value = None
                try:
                    value = func(*args, **kwargs)
                except Exception as exc:
                    request: Request = args[0]
                    context = request.state.context
                    api: sly.Api = request.state.api
                    track_id = context["trackId"]
                    api.logger.error("An error occured:")
                    api.logger.exception(exc)

                    api.post(
                        "videos.notify-annotation-tool",
                        data={
                            "type": "videos:tracking-error",
                            "data": {
                                "trackId": track_id,
                                "error": {"message": repr(exc)},
                            },
                        },
                    )
                return value

            return wrapper

        @send_error_data
        def track(request: Request):
            context = request.state.context
            api: sly.Api = request.state.api

            video_interface = TrackerInterface(
                context=context,
                api=api,
                load_all_frames=False,
                frame_loader=self.download_frame,
            )

            range_of_frames = [
                video_interface.frames_indexes[0],
                video_interface.frames_indexes[-1],
            ]

            self.run_cache_task_manually(
                api,
                [range_of_frames],
                video_id=video_interface.video_id,
            )
            api.logger.info("Start tracking.")

            for _ in video_interface.frames_loader_generator():
                for (fig_id, geom), obj_id in zip(
                    video_interface.geometries.items(),
                    video_interface.object_ids,
                ):
                    if isinstance(geom, sly.Point):
                        geometries = self._predict_point_geometries(
                            geom,
                            video_interface,
                        )
                    elif isinstance(geom, sly.Polygon):
                        if len(geom.interior) > 0:
                            raise ValueError("Can't track polygons with iterior.")
                        geometries = self._predict_polygon_geometries(
                            geom,
                            video_interface,
                        )
                    elif isinstance(geom, sly.GraphNodes):
                        geometries = self._predict_graph_geometries(
                            geom,
                            video_interface,
                        )
                    elif isinstance(geom, sly.Polyline):
                        geometries = self._predict_polyline_geometries(
                            geom,
                            video_interface,
                        )
                    else:
                        raise TypeError(f"Tracking does not work with {geom.geometry_name()}.")

                    video_interface.add_object_geometries(geometries, obj_id, fig_id)
                    api.logger.info(f"Object #{obj_id} tracked.")

                    if video_interface.global_stop_indicatior:
                        return

    def predict(
        self,
        rgb_images: List[np.ndarray],
        settings: Dict[str, Any],
        start_object: PredictionPoint,
    ) -> List[PredictionPoint]:
        """
        Track point on given frames.

        :param rgb_images: RGB frames, `m` frames
        :type rgb_images: List[np.array]
        :param settings: model parameters
        :type settings: Dict[str, Any]
        :param start_objects: point to track on the initial frame
        :type start_objects: PredictionPoint
        :return: predicted points for frame range (0, m]; `m-1` prediction in total
        :rtype: List[PredictionPoint]
        """
        raise NotImplementedError

    def visualize(
        self,
        predictions: List[PredictionPoint],
        images: List[np.ndarray],
        vis_path: str,
        thickness: int = 2,
    ):
        vis_path = Path(vis_path)

        for i, (pred, image) in enumerate(zip(predictions, images)):
            out_path = vis_path / f"img_{i}.jpg"
            ann = self._predictions_to_annotation(image, [pred])
            ann.draw_pretty(
                bitmap=image,
                color=(255, 0, 0),
                thickness=thickness,
                output_path=str(out_path),
                fill_rectangles=False,
            )

    def _create_label(self, dto: PredictionPoint) -> sly.Point:
        geometry = sly.Point(row=dto.row, col=dto.col)
        return Label(geometry, sly.ObjClass("", sly.Point))

    def _get_obj_class_shape(self):
        return sly.Point

    def _predict_point_geometries(
        self,
        geom: sly.Point,
        interface: TrackerInterface,
    ) -> List[sly.Point]:
        pp_geom = PredictionPoint("point", col=geom.col, row=geom.row)
        predicted: List[Prediction] = self.predict(
            interface.frames_with_notification,
            self.custom_inference_settings_dict,
            pp_geom,
        )
        return F.dto_points_to_sly_points(predicted)

    def _predict_polygon_geometries(
        self,
        geom: sly.Polygon,
        interface: TrackerInterface,
    ) -> List[sly.Polygon]:
        polygon_points = F.numpy_to_dto_point(geom.exterior_np, "polygon")
        exterior_per_time = [[] for _ in range(interface.frames_count)]

        for pp_geom in polygon_points:
            points: List[Prediction] = self.predict(
                interface.frames_with_notification,
                self.custom_inference_settings_dict,
                pp_geom,
            )
            points_loc = F.dto_points_to_point_location(points)
            for fi, point_loc in enumerate(points_loc):
                exterior_per_time[fi].append(point_loc)

        return F.exteriors_to_sly_polygons(exterior_per_time)

    def _predict_graph_geometries(
        self,
        geom: sly.GraphNodes,
        interface: TrackerInterface,
    ) -> sly.GraphNodes:
        nodes_per_time = [[] for _ in range(interface.frames_count)]
        points_with_id = F.graph_to_dto_points(geom)

        for point, pid in zip(*points_with_id):
            preds: List[PredictionPoint] = self.predict(
                interface.frames_with_notification,
                self.custom_inference_settings_dict,
                point,
            )
            nodes = F.dto_points_to_sly_nodes(preds, pid)

            for time, node in enumerate(nodes):
                nodes_per_time[time].append(node)

        return F.nodes_to_sly_graph(nodes_per_time)

    def _predict_polyline_geometries(
        self,
        geom: sly.Polyline,
        interface: TrackerInterface,
    ) -> List[sly.Polyline]:
        polyline_points = F.numpy_to_dto_point(geom.exterior_np, "polyline")
        lines_per_time = [[] for _ in range(interface.frames_count)]

        for point in polyline_points:
            preds: List[PredictionPoint] = self.predict(
                interface.frames_with_notification,
                self.custom_inference_settings_dict,
                point,
            )
            sly_points_loc = F.dto_points_to_point_location(preds)

            for time, point_loc in enumerate(sly_points_loc):
                lines_per_time[time].append(point_loc)

        return F.exterior_to_sly_polyline(lines_per_time)

    def _predictions_to_annotation(
        self, image: np.ndarray, predictions: List[Prediction]
    ) -> sly.Annotation:
        labels = []
        for prediction in predictions:
            label = self._create_label(prediction)
            if label is None:
                # for example empty mask
                continue
            if isinstance(label, list):
                labels.extend(label)
                continue
            labels.append(label)

        # create annotation with correct image resolution
        ann = sly.Annotation(img_size=image.shape[:2])
        ann = ann.add_labels(labels)
        return ann



# coding: utf-8

from __future__ import annotations
from typing import List, Dict, Optional
import uuid
from bidict import bidict

from supervisely.annotation.label import LabelJsonFields
from supervisely.annotation.obj_class import ObjClass
from supervisely.project.project_meta import ProjectMeta
from supervisely._utils import take_with_default
from supervisely.video_annotation.constants import KEY, ID, OBJECTS_MAP
from supervisely.video_annotation.video_tag_collection import VideoTagCollection
from supervisely.video_annotation.video_tag import VideoTag
from supervisely.collection.key_indexed_collection import KeyObject
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.geometry.constants import (
    LABELER_LOGIN,
    UPDATED_AT,
    CREATED_AT,
    CLASS_ID,
)


class VideoObject(KeyObject):
    """
    VideoObject object for :class:`VideoAnnotation<supervisely.video_annotation.video_annotation.VideoAnnotation>`. :class:`VideoObject<VideoObject>` object is immutable.

    :param obj_class: VideoObject :class:`class<supervisely.annotation.obj_class.ObjClass>`.
    :type obj_class: ObjClass
    :param tags: VideoObject :class:`tags<supervisely.video_annotation.video_tag_collection.VideoTagCollection>`.
    :type tags: VideoTagCollection, optional
    :param key: KeyIdMap object.
    :type key: KeyIdMap, optional
    :param class_id: ID of :class:`ObjClass<supervisely.annotation.obj_class.ObjClass>` to which VideoObject belongs.
    :type class_id: int, optional
    :param labeler_login: Login of the user who created VideoObject.
    :type labeler_login: str, optional
    :param updated_at: Date and Time when VideoObject was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
    :type updated_at: str, optional
    :param created_at: Date and Time when VideoObject was created. Date Format is the same as in "updated_at" parameter.
    :type created_at: str, optional
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        obj_class_car = sly.ObjClass('car', sly.Rectangle)
        video_obj_car = sly.VideoObject(obj_class_car)
        video_obj_car_json = video_obj_car.to_json()
        print(video_obj_car_json)
        # Output: {
        #     "key": "6b819f1840f84d669b32cdec225385f0",
        #     "classTitle": "car",
        #     "tags": []
        # }
    """
    def __init__(self, obj_class: ObjClass, tags: Optional[VideoTagCollection] = None, key: Optional[uuid.UUID]=None,
                 class_id: Optional[int]=None, labeler_login: Optional[str]=None, updated_at: Optional[str]=None,
                 created_at: Optional[str]=None):
        self.labeler_login = labeler_login
        self.updated_at = updated_at
        self.created_at = created_at

        self._obj_class = obj_class
        self._key = take_with_default(key, uuid.uuid4())
        self._tags = take_with_default(tags, VideoTagCollection())
        self._class_id = take_with_default(class_id, None)

    def _add_creation_info(self, d):
        if self.labeler_login is not None:
            d[LABELER_LOGIN] = self.labeler_login
        if self.updated_at is not None:
            d[UPDATED_AT] = self.updated_at
        if self.created_at is not None:
            d[CREATED_AT] = self.created_at

    @property
    def obj_class(self) -> ObjClass:
        """
        ObjClass of the current VideoObject.

        :return: ObjClass object
        :rtype: :class:`ObjClass<supervisely.annotation.obj_class.ObjClass>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            obj_class_car = sly.ObjClass('car', sly.Rectangle)
            video_obj_car = sly.VideoObject(obj_class_car)
            obj_car_json = video_obj_car.obj_class.to_json()
            print(obj_car_json)
            # Output: {
            #     "title": "car",
            #     "shape": "rectangle",
            #     "color": "#8A0F7B",
            #     "geometry_config": {},
            #     "hotkey": ""
            # }
        """
        return self._obj_class

    def key(self) -> uuid.UUID:
        """
        Object key.

        :return: Object key
        :rtype: uuid.UUID
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            obj_class_car = sly.ObjClass('car', sly.Rectangle)
            video_obj_car = sly.VideoObject(obj_class_car)
            key = video_obj_car.key()
            print(key)
            # Output: 158e6cf4f4ac4c639fc6994aad127c16
        """

        return self._key

    @property
    def tags(self) -> VideoTagCollection:
        """
        VideoTagCollection of the current VideoObject.

        :return: VideoTagCollection object
        :rtype: :class:`VideoTagCollection<supervisely.video_annotation.video_tag_collection.VideoTagCollection>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            # Create VideoTagCollection
            meta_car = sly.TagMeta('car_tag', sly.TagValueType.ANY_STRING)
            from supervisely.video_annotation.video_tag import VideoTag
            car_tag = VideoTag(meta_car, value='acura')
            from supervisely.video_annotation.video_tag_collection import VideoTagCollection
            video_tags = VideoTagCollection([car_tag])

            # Create VideoObject
            obj_class_car = sly.ObjClass('car', sly.Rectangle)
            video_obj_car = sly.VideoObject(obj_class_car, video_tags)

            # Get VideoTagCollection
            tags = video_obj_car.tags
            tags_json = tags.to_json()
            print(tags_json)
            # Output: [
            #     {
            #         "name": "car_tag",
            #         "value": "acura",
            #         "key": "4f82fbcab74c44259d7a0e29d604602e"
            #     }
            # ]
        """
        return self._tags.clone()

    @property
    def class_id(self) -> int:
        """
        Object class ID.

        :return: Object class ID.
        :rtype: int
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            obj_class_car = sly.ObjClass('car', sly.Rectangle)
            video_obj_car = sly.VideoObject(obj_class_car)
            class_id = video_obj_car.class_id
        """

        return self._class_id

    def add_tag(self, tag: VideoTag) -> VideoObject:
        """
        Adds VideoTag to the current VideoObject.

        :param tag: VideoTag to be added.
        :type tag: VideoTag
        :return: VideoObject object
        :rtype: :class:`VideoObject<VideoObject>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            # Create VideoObject
            obj_class_car = sly.ObjClass('car', sly.Rectangle)
            video_obj_car = sly.VideoObject(obj_class_car)

            # Create VideoTag
            meta_car = sly.TagMeta('car_tag', sly.TagValueType.ANY_STRING)
            from supervisely.video_annotation.video_tag import VideoTag
            car_tag = VideoTag(meta_car, value='acura')

            # Add VideoTag
            new_obj_car = video_obj_car.add_tag(car_tag)
            new_obj_car_json = new_obj_car.to_json()
            print(new_obj_car_json)
            # Output: {
            #     "key": "1ab52285ee634c93b724fa655b785eae",
            #     "classTitle": "car",
            #     "tags": [
            #         {
            #             "name": "car_tag",
            #             "value": "acura",
            #             "key": "d9e52b275e074c538f162a6d679aed9e"
            #         }
            #     ]
            # }
        """
        return self.clone(tags=self._tags.add(tag))

    def add_tags(self, tags: List[VideoTag]) -> VideoObject:
        """
        Adds VideoTags to the current VideoObject.

        :param tag: List of VideoTags to be added.
        :type tag: List[VideoTag]
        :return: VideoObject object
        :rtype: :class:`VideoObject<VideoObject>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            # Create VideoObject
            obj_class_vehicle = sly.ObjClass('vehicle', sly.Rectangle)
            video_obj_vehicle = sly.VideoObject(obj_class_vehicle)

            # Create VideoTags
            from supervisely.video_annotation.video_tag import VideoTag
            meta_car = sly.TagMeta('car_tag', sly.TagValueType.ANY_STRING)
            car_tag = VideoTag(meta_car, value='acura')
            meta_bus = sly.TagMeta('bus_tag', sly.TagValueType.ANY_STRING)
            bus_tag = VideoTag(meta_bus, value='volvo')

            # Add VideoTags
            new_obj_vehicle = video_obj_vehicle.add_tags([car_tag, bus_tag])
            new_obj_vehicle_json = new_obj_vehicle.to_json()
            print(new_obj_vehicle_json)
            # Output: {
            #     "key": "94055c5e8cb146368f627fc608fb6b44",
            #     "classTitle": "vehicle",
            #     "tags": [
            #         {
            #             "name": "car_tag",
            #             "value": "acura",
            #             "key": "6679f47c96734565919fbffc278532a1"
            #         },
            #         {
            #             "name": "bus_tag",
            #             "value": "volvo",
            #             "key": "d0a60dc929de491a85a09fea59adb818"
            #         }
            #     ]
            # }
        """
        return self.clone(tags=self._tags.add_items(tags))

    def to_json(self, key_id_map: Optional[KeyIdMap] = None) -> Dict:
        """
        Convert the VideoObject to a json dict. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :return: Json format as a dict
        :rtype: :class:`dict`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            obj_class_car = sly.ObjClass('vehicle', sly.Rectangle)
            video_obj_car = sly.VideoObject(obj_class_car)

            obj_car_json = video_obj_car.to_json()
            print(obj_car_json)
            # Output: {
            #     "key": "ce26e77a45bc45e88e3e17da1672d01f",
            #     "classTitle": "vehicle",
            #     "tags": []
            # }
        """
        data_json = {
            KEY: self.key().hex,
            LabelJsonFields.OBJ_CLASS_NAME: self.obj_class.name,
            LabelJsonFields.TAGS: self.tags.to_json(key_id_map),
        }

        if key_id_map is not None:
            item_id = key_id_map.get_object_id(self.key())
            if item_id is not None:
                data_json[ID] = item_id

        self._add_creation_info(data_json)
        return data_json

    @classmethod
    def from_json(cls, data: Dict, project_meta: ProjectMeta, key_id_map: Optional[KeyIdMap] = None) -> VideoObject:
        """
        Convert a json dict to VideoObject. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :param data: Dict in json format.
        :type data: dict
        :param project_meta: Input :class:`ProjectMeta<supervisely.project.project_meta.ProjectMeta>`.
        :type project_meta: ProjectMeta
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :raises: :class:`RuntimeError`, if object class name is not found in the given project meta
        :return: VideoObject object
        :rtype: :class:`VideoObject`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            obj_car_json = {
                "classTitle": "vehicle",
                "tags": []
            }
            obj_class_car = sly.ObjClass('vehicle', sly.Rectangle)
            obj_collection = sly.ObjClassCollection([obj_class_car])
            meta = sly.ProjectMeta(obj_classes=obj_collection)

            video_obj_from_json = sly.VideoObject.from_json(obj_car_json, meta)
        """
        obj_class_name = data[LabelJsonFields.OBJ_CLASS_NAME]
        obj_class = project_meta.get_obj_class(obj_class_name)
        if obj_class is None:
            raise RuntimeError(
                f"Failed to deserialize a object from JSON: class name {obj_class_name!r} "
                f"was not found in the given project meta."
            )

        key = uuid.UUID(data[KEY]) if KEY in data else uuid.uuid4()

        if key_id_map is not None:
            key_id_map.add_object(key, data.get(ID, None))

        class_id = data.get(CLASS_ID, None)
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)

        return cls(
            obj_class=obj_class,
            key=key,
            tags=VideoTagCollection.from_json(
                data[LabelJsonFields.TAGS], project_meta.tag_metas
            ),
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )

    def clone(self, obj_class: Optional[ObjClass]=None, tags: Optional[VideoTagCollection] = None, key: Optional[KeyIdMap]=None,
              class_id: Optional[int]=None, labeler_login: Optional[str]=None, updated_at: Optional[str]=None, created_at: Optional[str]=None) -> VideoObject:
        """
        Makes a copy of VideoObject with new fields, if fields are given, otherwise it will use fields of the original VideoObject.

        :param obj_class: VideoObject :class:`class<supervisely.annotation.obj_class.ObjClass>`.
        :type obj_class: ObjClass, optional
        :param tags: VideoObject :class:`tags<supervisely.video_annotation.video_tag_collection.VideoTagCollection>`.
        :type tags: VideoTagCollection, optional
        :param key: KeyIdMap object.
        :type key: KeyIdMap, optional
        :param class_id: ID of :class:`ObjClass<supervisely.annotation.obj_class.ObjClass>` to which VideoObject belongs.
        :type class_id: int, optional
        :param labeler_login: Login of the user who created VideoObject.
        :type labeler_login: str, optional
        :param updated_at: Date and Time when VideoObject was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
        :type updated_at: str, optional
        :param created_at: Date and Time when VideoObject was created. Date Format is the same as in "updated_at" parameter.
        :type created_at: str, optional
        :return: VideoObject object
        :rtype: :class:`VideoObject`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            obj_class_car = sly.ObjClass('vehicle', sly.Rectangle)
            video_obj_car = sly.VideoObject(obj_class_car)

            # New ObjClass to clone
            obj_class_bus = sly.ObjClass('bus', sly.Rectangle)

            # New VideoTagCollection to clone
            from supervisely.video_annotation.video_tag import VideoTag
            meta_car = sly.TagMeta('car_tag', sly.TagValueType.ANY_STRING)
            car_tag = VideoTag(meta_car, value='acura')
            from supervisely.video_annotation.video_tag_collection import VideoTagCollection
            tags = VideoTagCollection([car_tag])
            # Clone
            new_obj_vehicle = video_obj_car.clone(obj_class=obj_class_bus, tags=tags)
            new_obj_vehicle_json = new_obj_vehicle.to_json()
            print(new_obj_vehicle_json)
            # Output: {
            #     "key": "39ae5b9ce1ca405c9f53544374b3f5be",
            #     "classTitle": "bus",
            #     "tags": [
            #         {
            #             "name": "car_tag",
            #             "value": "acura",
            #             "key": "3119b6e38fe24fe7a220e881154fd9ba"
            #         }
            #     ]
            # }
        """
        return self.__class__(obj_class=take_with_default(obj_class, self.obj_class),
                              key=take_with_default(key, self._key),
                              tags=take_with_default(tags, self.tags),
                              class_id=take_with_default(class_id, self.class_id),
                              labeler_login=take_with_default(labeler_login, self.labeler_login),
                              updated_at=take_with_default(updated_at, self.updated_at),
                              created_at=take_with_default(created_at, self.created_at))




import sys
import click


from supervisely.cli.project import download_run, get_project_name_run
from supervisely.cli.task import set_output_directory_run
from supervisely.cli.teamfiles import (
    remove_file_run,
    remove_directory_run,
    upload_directory_run,
    download_directory_run,
)


@click.group()
def cli():
    pass


@cli.command(help="This app allows you to release your aplication to Supervisely platform")
@click.option(
    "-p",
    "--path",
    required=False,
    help="[Optional] Path to the directory with application",
)
@click.option(
    "-a",
    "--sub-app",
    required=False,
    help="[Optional] Path to sub-app relative to application directory",
)
@click.option(
    "--release-version",
    required=False,
    help='[Optional] Release version in format "vX.X.X"',
)
@click.option(
    "--release-description",
    required=False,
    help="[Optional] Release description (max length is 64 symbols)",
)
@click.option(
    "--share",
    is_flag=True,
    help="[Optional] Add this flag to share the private app on your private instance",
)
@click.option("-y", is_flag=True, help="[Optional] Add this flag for autoconfirm")
@click.option("-s", "--slug", required=False, help="[Optional] For internal use")
def release(path, sub_app, slug, y, release_version, release_description, share):
    from supervisely.cli.release import run

    try:
        success = run(
            app_directory=path,
            sub_app_directory=sub_app,
            slug=slug,
            autoconfirm=y,
            release_version=release_version,
            release_description=release_description,
            share=share,
        )
        if success:
            print("App released sucessfully!")
            sys.exit(0)
        else:
            print("App not released")
            sys.exit(1)
    except KeyboardInterrupt:
        print("Aborting...")
        print("App not released")
        sys.exit(1)


@cli.group()
def project():
    """Commands: download, get-name"""
    pass


@project.command(help="Download project data from supervisely to local directory")
@click.option(
    "-id",
    "--id",
    required=True,
    type=int,
    help="Supervisely project ID",
)
@click.option(
    "-d",
    "--dst",
    required=True,
    type=str,
    help="Download destination directory",
)
def download(id: int, dst: str) -> None:
    try:
        success = download_run(id, dst)
        if success:
            sys.exit(0)
        else:
            sys.exit(1)
    except KeyboardInterrupt:
        print("\nDownload aborted\n")
        sys.exit(1)


@project.command(help="Get project name")
@click.option(
    "-id",
    "--id",
    required=True,
    type=int,
    help="Supervisely project ID",
)
def get_name(id: int) -> None:
    try:
        success = get_project_name_run(id)
        if success:
            sys.exit(0)
        else:
            print("Getting project name failed")
            sys.exit(1)
    except KeyboardInterrupt:
        print("Getting project name directory aborted")
        sys.exit(1)


@cli.group()
def teamfiles():
    """Commands: download, remove-file, remove-dir, upload"""
    pass


@teamfiles.command(
    help="Download source files from Team files directory with destination to local path"
)
@click.option(
    "-id",
    "--id",
    required=True,
    type=int,
    help="Supervisely team ID",
)
@click.option(
    "-s",
    "--src",
    required=True,
    type=str,
    help="Path to Team files source directory from which files are downloaded",
)
@click.option(
    "-d",
    "--dst",
    required=True,
    type=str,
    help="Path to local destination directory to which files are downloaded",
)
@click.option(
    "-f",
    "--filter",
    required=False,
    type=str,
    help="[Optional] Filter downloaded files using regexp",
)
@click.option(
    "-i",
    is_flag=True,
    help="[Optional] Ignore and skip if source directory not exists",
)
def download(id: int, src: str, dst: str, filter: str, i: bool) -> None:
    try:
        success = download_directory_run(id, src, dst, filter, ignore_if_not_exists=i)
        if success:
            sys.exit(0)
        else:
            sys.exit(1)
    except KeyboardInterrupt:
        print("Download aborted")
        sys.exit(1)


@teamfiles.command(help="Remove file from supervisely teamfiles")
@click.option(
    "-id",
    "--id",
    required=True,
    type=int,
    help="Supervisely team ID",
)
@click.option(
    "-p",
    "--path",
    required=True,
    type=str,
    help="File path to remove",
)
def remove_file(id: int, path: str) -> None:
    try:
        success = remove_file_run(id, path)
        if success:
            sys.exit(0)
        else:
            sys.exit(1)
    except KeyboardInterrupt:
        print("\nRemoving file aborted\n")
        sys.exit(1)


@teamfiles.command(help="Remove directory from supervisely Team files")
@click.option(
    "-id",
    "--id",
    required=True,
    type=int,
    help="Supervisely team ID",
)
@click.option(
    "-p",
    "--path",
    required=True,
    type=str,
    help="Path to remove directory",
)
def remove_dir(id: int, path: str) -> None:
    try:
        success = remove_directory_run(id, path)
        if success:
            sys.exit(0)
        else:
            sys.exit(1)
    except KeyboardInterrupt:
        print("Removing directory aborted")
        sys.exit(1)


@teamfiles.command(help="Upload local source files with destination to supervisely Team files")
@click.option(
    "-id",
    "--id",
    required=True,
    type=int,
    help="Supervisely team ID",
)
@click.option(
    "-s",
    "--src",
    required=True,
    type=str,
    help="Path to local source directory from which files are uploaded",
)
@click.option(
    "-d",
    "--dst",
    required=True,
    type=str,
    help="Path to Team files remote destination directory to which files are uploaded",
)
def upload(id: int, src: str, dst: str) -> None:
    try:
        success = upload_directory_run(id, src, dst)
        if success:
            sys.exit(0)
        else:
            sys.exit(1)
    except KeyboardInterrupt:
        print("Upload aborted")
        sys.exit(1)


@cli.group()
def task():
    """Commands: set-output-dir"""
    pass


@task.command(help="Set link to Team files directory at workspace tasks interface")
@click.option(
    "-id",
    "--id",
    required=False,
    type=int,
    help="[Optional] Supervisely task ID",
)
@click.option(
    "--team-id",
    required=False,
    type=int,
    help="[Optional] Supervisely team ID",
)
@click.option(
    "-d",
    "--dir",
    required=True,
    type=str,
    help="Path to Team files directory",
)
def set_output_dir(id: int, team_id: int, dir: str) -> None:
    try:
        success = set_output_directory_run(id, team_id, dir)
        if success:
            sys.exit(0)
        else:
            sys.exit(1)
    except KeyboardInterrupt:
        print("Setting task output directory aborted")
        sys.exit(1)



from typing import Union
from typing import NamedTuple, List, Dict
from supervisely.app.widgets.apexchart.apexchart import Apexchart
from supervisely.app.content import DataJson


class TreemapChart(Apexchart):
    """Widget for displaying data for series in a comparison by distribution of each of them.
    Supports multiple series, which can be added in one chart. Widget also handles click events on datapoints.

    :param title: title of the chart, which will be displayed above the chart
    :type title: str
    :param colors: list of colors for series, if not specified, default color will be used,
        if there are more series than colors, colors will be repeated. colors should be in hex format (e.g. #008FFB)
    :type colors: List[str]
    :param tooltip: tooltip for chart, which will be displayed on hover over datapoint
    :type tooltip: str

    :Methods:
    add_series(): add new series to the chart, if len(names) != len(values), ValueError will be raised
    set_series(): set series to the chart, deleting all previous series, if len(names) != len(values),
        ValueError will be raised
    get_series(): get series by index. If index is out of range, IndexError will be raised
    delete_series(): delete series by index. If index is out of range, IndexError will be raised
    @click(): decorator for handling click events on datapoints. It will return ClickedDataPoint object.

    :Usage example:

     .. code-block:: python
        from supervisely.app.widgets import TreemapChart

        # Preparing list of custom colors for series (it's optional)
        colors = [
            "#008FFB",
            "#00E396",
            "#FEB019",
            "#FF4560",
            "#775DD0",
            ]

        # Creating tooltip for chart (it's optional)
        tooltip = "There are {y} objects"

        # Initialize empty chart without any series.
        tc = TreemapChart(title="Treemap Chart", colors=colors)


        # Preparing series data for widget
        # Remember that names and values must have the same length, otherwise ValueError will be raised
        names = ["cats", "dogs", "birds", "fishes", "snakes"]
        values = [3, 5, 1, 2, 1]

        # Setting series to the chart (aware that it will delete all previous series if they existed)
        dc.set_series(names, values)

    """

    class ClickedDataPoint(NamedTuple):
        """Class, representing clicked datapoint, which contains information about series, data index and data itself.
        It will be returned after click event on datapoint in unmutable namedtuple
        with fields: series_index, data_index, data."""

        series_index: int
        data_index: int
        data: dict

    def __init__(
        self,
        title: str,
        colors: List[str] = None,
        tooltip: str = None,
    ):
        self._title = title
        self._series = []
        self._widget_height = 350
        self._tooltip = tooltip

        if not colors:
            self._distributed = False
            self._colors = ["#008FFB"]
        else:
            self._distributed = True
            self._colors = colors

        self._options = {
            "chart": {
                "type": "treemap",
            },
            "legend": {
                "show": False,
            },
            "plotOptions": {
                "treemap": {
                    "distributed": self._distributed,
                    "enableShades": False,
                }
            },
            "title": {"text": self._title, "align": "left"},
            "colors": self._colors,
        }

        sly_options = {}
        if self._tooltip is not None:
            sly_options["tooltip"] = self._tooltip
            self._options["tooltip"] = {"y": {}}

        super(TreemapChart, self).__init__(
            series=self._series,
            options=self._options,
            type="treemap",
            height=self._widget_height,
            sly_options=sly_options,
        )

    def _manage_series(self, names: List[str], values: List[Union[int, float]], set: bool = False):
        """This is a private method, which should not be used directly. Use add_series() or set_series() instead.
        It will add or set series to the chart, depending on set parameter. If set is True, all previous series will be
        deleted, otherwise new series will be added to the chart.

        :param names: list of names of the cells in series, which will be displayed on the chart
        :type names: List[str]
        :param values: list of values of the cells in series, which will be displayed on the chart
        :type values: List[Union[int, float]
        :param set: if True, all previous series will be deleted, otherwise new series will be added to the chart,
            defaults to False
        :type set: bool, optional
        :raises ValueError: if names and values has different length
        :raises ValueError: if any of values is not int or float
        """
        if len(names) != len(values):
            raise ValueError(
                f"Names and values has different length: {len(names)} != {len(values)}"
            )
        for value in values:
            if not isinstance(value, (int, float)):
                raise ValueError(
                    f"All values must be ints or floats, but {value} is {type(value)}."
                )

        data = [{"x": x, "y": y} for x, y in zip(names, values)]
        if set:
            self._series = [{"data": data}]
        else:
            self._series.append({"data": data})
        self.update_data()
        DataJson().send_changes()

    def add_series(self, names: List[str], values: List[Union[int, float]]):
        """Adds new series to the chart. Len of names and values must be equal, otherwise ValueError will be raised.

        :param names: list of names of the cells in series, which will be displayed on the chart
        :type names: List[str]
        :param values: list of values of the cells in series, which will be displayed on the chart
        :type values: List[Union[int, float]]
        """
        self._manage_series(names, values)

    def set_series(self, names: List[str], values: List[Union[int, float]]):
        """Sets series to the chart, deleting all previous series. Len of names and values must be equal,
        otherwise ValueError will be raised.

        :param names: list of names of the cells in series, which will be displayed on the chart
        :type names: List[str]
        :param values: list of values of the cells in series, which will be displayed on the chart
        :type values: List[Union[int, float]]
        """
        self._manage_series(names, values, set=True)

    def get_series(self, index: int) -> Dict[str, List[Dict[str, Union[int, float]]]]:
        """Returns series by index. If index is out of range, IndexError will be raised.
        Returned series is a dict with key "data" and a value is a list of dicts with keys "x" and "y",
        where "x" is a name of the cell and "y" is a value of the cell.

        :param index: index of the series, if index is out of range, IndexError will be raised
        :type index: int
        :raises TypeError: if index is not int
        :raises IndexError: if index is out of range
        :return: series data by given index
        :rtype: Dict[str, Union[int, float]]
        """

        if not isinstance(index, int):
            raise TypeError(f"Index must be int, but {index} is {type(index)}.")
        try:
            return self._series[index]
        except IndexError:
            raise IndexError(f"Series with index {index} does not exist.")

    def delete_series(self, index: int):
        """Removes series by index from the chart. If index is out of range, IndexError will be raised.

        :param index: index of the series to delete
        :type index: int
        :raises TypeError: if index is not int
        :raises IndexError: if index is out of range
        """
        if not isinstance(index, int):
            raise TypeError(f"Index must be int, but {index} is {type(index)}.")
        try:
            del self._series[index]
        except IndexError:
            raise IndexError(f"Series with index {index} does not exist.")
        self.update_data()
        DataJson().send_changes()

    def get_clicked_datapoint(self) -> Union[ClickedDataPoint, None]:
        """Returns clicked datapoint as a ClickedDataPoint object, which is a namedtuple with fields:
        series_index, data_index and data. If click was outside of the cells, None will be returned.

        :return: clicked datapoint as a ClickedDataPoint object or None if click was outside of the cells
        :rtype: Union[ClickedDataPoint, None]
        """
        value = self.get_clicked_value()
        series_index = value["seriesIndex"]
        data_index = value["dataPointIndex"]

        if series_index == -1 or data_index == -1:
            # If click was outside of the cells.
            return

        raw_data = self._series[series_index]["data"][data_index]
        data = {
            "name": raw_data["x"],
            "value": raw_data["y"],
        }

        res = TreemapChart.ClickedDataPoint(series_index, data_index, data)

        return res
